/** @file rob_observability.cpp verification of observability algorithm */

/*
 * Implementation of verification of observability algorithm introduced in 
 * C.G. Cassandras, S. Lafortune: Introduction to Discrete Event Systems, 
 * 2nd edition
 * 
 * Algorithm verifies if K is observable with respect to M, Eo and Ee.
 * Input of a test:
 * L(G) = M.
 * L(H) = prefix closure of K, K is subset of M
 * Eo is subset of E; Euo = E - Eo. Let P: E* -> Eo*.
 * Ec is subset of E.
 * 
 * Output of a test:
 * ObsTest(H, G) - deterministic generator with transition function
 * involving event triples. Only marked state is state "dead".
 * 
 * If Lm(ObsTest(H, G) is empty set then K is observable with respect to M, Eo and Ee.
 * 
 * Implemented by Marian Pochyba, 2018  
 */

#include <string>
#include <queue>

#include "rob_observability.h"

namespace faudes {

// struct to represent state triple with each index
typedef struct StateTriple {
    Idx x1;
    Idx x2;
    Idx x3;
} StateTriple;

// struct to represent event triple with each index
typedef struct EventTriple {
    Idx x1;
    Idx x2;
    Idx x3;
} EventTriple;

// create initial state triple for the construction of ObsTest(H, G)
StateTriple getInitState(const Generator &h, const Generator &g) {

    StateTriple initState;

    Idx hInitIndex = h.InitState();
    Idx gInitIndex = g.InitState();

    initState.x1 = hInitIndex;
    initState.x2 = hInitIndex;
    initState.x3 = gInitIndex;

    return initState;
}

// simply just create EventTriple variable from 3 indices
EventTriple createEventTriple(Idx x1, Idx x2, Idx x3) {

    EventTriple event;
    event.x1 = x1;
    event.x2 = x2;
    event.x3 = x3;

    return event;

}

// returns name representation of state in ObsTest(H, G)
// based on names of states in generators H and G
std::string getStateTripleName(const StateTriple &state,
                               const Generator &h,
                               const Generator &g) {

    return h.StateName(state.x1) + "/" + h.StateName(state.x2) + "/" + g.StateName(state.x3);
}

// returns name representation of event in ObsTest(H,G)
// based on names of events in generators H and G
std::string getEventTripleName(const EventTriple &event,
                               const Generator &h,
                               const Generator &g) {

    return h.EventName(event.x1) + "/" + h.EventName(event.x2) + "/" + g.EventName(event.x3);
}

// returns state triple in ObsTest(H, G) based on current state triple,
// event triple and transitions in generators H and G
StateTriple getNextState(const StateTriple &state,
                         const EventTriple &event,
                         const Generator &h,
                         const Generator &g) {

    StateTriple nextState;
    nextState.x1 = *h.ActiveTransSet(state.x1).SuccessorStates(state.x1, event.x1).Begin();
    nextState.x2 = *h.ActiveTransSet(state.x2).SuccessorStates(state.x2, event.x2).Begin();
    nextState.x3 = *g.ActiveTransSet(state.x3).SuccessorStates(state.x3, event.x3).Begin();

    return nextState;
}

// returns active event set of state triple in ObsTest(H, G)
// based on active event sets of fragment states in generators H and G
EventSet getActEventSet(const StateTriple &state,
                        const Generator &h,
                        const Generator &g) {

    EventSet ActEventSet;
    ActEventSet = h.ActiveEventSet(state.x1)
                + h.ActiveEventSet(state.x2)
                + g.ActiveEventSet(state.x3);

    return ActEventSet;
}

// adds epsilon self loops into a generator transition set
void addEpsilonSelfLoops(Generator &gen) {

    StateSet::Iterator it;
    for (it = gen.StatesBegin(); it != gen.StatesEnd(); ++it) {
        gen.SetTransition(*it, gen.EventIndex("_"), *it);
    }
}

// set ups generators, epsilon index and queue before algorithm
void setUp(Generator &h,
           Generator &g,
           Idx &epsilonIdx,
           std::queue<StateTriple> &queue) {

    h.InsEvent("_");
    addEpsilonSelfLoops(h);
    g.InsEvent("_");
    addEpsilonSelfLoops(g);

    StateSet::Iterator stateIt;
    for (stateIt = h.StatesBegin(); stateIt != h.StatesEnd(); ++stateIt) {
        if (h.StateName(*stateIt).empty()) {
            std::string name = static_cast<std::ostringstream*>( &(std::ostringstream() << *stateIt) )->str();
            h.StateName(*stateIt, name);
        }
    }

    for (stateIt = g.StatesBegin(); stateIt != g.StatesEnd(); ++stateIt) {
        if (g.StateName(*stateIt).empty()) {
            std::string name = static_cast<std::ostringstream*>( &(std::ostringstream() << *stateIt) )->str();
            g.StateName(*stateIt, name);
        }
    }

    epsilonIdx = h.EventIndex("_");

    StateTriple initState = getInitState(h, g);
    queue.push(initState);
}

// creates and pushes new state triple into queue
void newState(const Generator &h,
              const Generator &g,
              const StateTriple &state,
              const EventTriple &event,
              std::queue<StateTriple> &queue) {

    StateTriple newState = getNextState(state, event, h, g);
    queue.push(newState);
}

// checks input conditions
void checkConditions(Generator h,
                     Generator g,
                     const EventSet &obsEvents,
                     const EventSet &conEvents) {

    if (!IsDeterministic(h)) {
        std::stringstream errstr;
        errstr << "Generator " << h.Name() << " not deterministic";
        throw Exception("Observability", errstr.str(), 101);
    }

    if (!IsDeterministic(g)) {
        std::stringstream errstr;
        errstr << "Generator " << g.Name() << " not deterministic";
        throw Exception("Observability", errstr.str(), 101);
    }

    MarkAllStates(g);
    MarkAllStates(h);
    if (!LanguageInclusion(h, g)) {
        std::stringstream errstr;
        errstr << "Lm(H) = K is not a subset of L(G) = M!";
        throw Exception("Observability", errstr.str(), 201);
    }

    EventSet sigma = g.Alphabet();
    if (!(obsEvents <= sigma)) {
        std::stringstream errstr;
        errstr << "Observable event set is not a subset of Alphabet of G!";
        throw Exception("Observability", errstr.str(), 202);
    }

    if (!(conEvents <= sigma)) {
        std::stringstream errstr;
        errstr << "Controlable event set is not a subset of Alphabet of G!";
        throw Exception("Observability", errstr.str(), 202);
    }

}

// isObservable() function
bool isObservable(Generator h,
                  Generator g,
                  const EventSet &obsEvents,
                  const EventSet &conEvents) {

    // check input conditions
    checkConditions(h, g, obsEvents, conEvents);

    // set up
    Idx epsilonIdx;
    std::queue<StateTriple> stateQueue;
    StateTriple state;
    EventSet actEventSet;
    setUp(h, g, epsilonIdx, stateQueue);

    while (!stateQueue.empty()) {

        state = stateQueue.front();
        stateQueue.pop();

        // divide active event set of state triple
        // into observable and unobservable events
        actEventSet = getActEventSet(state, h, g);
        EventSet obsActEvents = actEventSet * obsEvents;
        EventSet unobsActEvents = actEventSet - obsActEvents;
        unobsActEvents.Erase(epsilonIdx);

        EventSet::Iterator it;
        Idx eventIdx;

        // process observable events
        for (it = obsActEvents.Begin(); it != obsActEvents.End(); ++it) {

            eventIdx = *it;
            if (h.ExistsTransition(state.x1, eventIdx)
                    && g.ExistsTransition(state.x3, eventIdx)) {
                if (conEvents.Exists(eventIdx)
                        &&!h.ExistsTransition(state.x2, eventIdx)) {
                    // fh(x1, e)! && fh(x2, e)not! && fg(x3, e)!
                    // dead state accessible -> K is not observable
                    return false;
                } else {
                    // fh(x1, e)! && fh(x2, e)! && fg(x3, e)!
                    EventTriple event = createEventTriple(eventIdx, eventIdx, eventIdx);
                    newState(h, g, state, event, stateQueue);
                }
            }
        }

        // process unobservable events
        for (it = unobsActEvents.Begin(); it != unobsActEvents.End(); ++it) {

            eventIdx = *it;
            if (h.ExistsTransition(state.x1, eventIdx)) {
                if (conEvents.Exists(eventIdx)
                        && !h.ExistsTransition(state.x2, eventIdx)
                        && g.ExistsTransition(state.x3, eventIdx)) {
                    // fh(x1, e)! && fh(x2, e)not! && fg(x3, e)!
                    // dead state accessible -> K is not observable
                    return false;
                }
                // fh(x1, e)!
                EventTriple event = createEventTriple(eventIdx, epsilonIdx, epsilonIdx);
                newState(h, g, state, event, stateQueue);
            }

            if (h.ExistsTransition(state.x2, eventIdx)
                    && g.ExistsTransition(state.x3, eventIdx)) {
                // fh(x2, e)! && fg(x3, e)!
                EventTriple event = createEventTriple(epsilonIdx, eventIdx, eventIdx);
                newState(h, g, state, event, stateQueue);
            }
        }
    }

    return true;

}

// set ups generators, epsilon index and queue before algorithm
void setUp(Generator &h,
           Generator &g,
           Generator &obsTest,
           Idx &epsilonIdx,
           std::queue<StateTriple> &queue) {

    h.InsEvent("_");
    addEpsilonSelfLoops(h);
    g.InsEvent("_");
    addEpsilonSelfLoops(g);

    StateSet::Iterator stateIt;
    for (stateIt = h.StatesBegin(); stateIt != h.StatesEnd(); ++stateIt) {
        if (h.StateName(*stateIt).empty()) {
            std::string name = static_cast<std::ostringstream*>( &(std::ostringstream() << *stateIt) )->str();
            h.StateName(*stateIt, name);
        }
    }

    for (stateIt = g.StatesBegin(); stateIt != g.StatesEnd(); ++stateIt) {
        if (g.StateName(*stateIt).empty()) {
            std::string name = static_cast<std::ostringstream*>( &(std::ostringstream() << *stateIt) )->str();
            g.StateName(*stateIt, name);
        }
    }

    epsilonIdx = h.EventIndex("_");

    StateTriple initState = getInitState(h, g);
    queue.push(initState);

    obsTest.Name("Observability test generator");
    obsTest.InsInitState(getStateTripleName(initState, h, g));
    obsTest.InsState("dead");
}

// adds new transition to state "dead" into transition set of ObsTest(H, G)
void newTransToDead(const Generator &h,
                    const Generator &g,
                    Generator &obsTest,
                    const StateTriple &state,
                    const EventTriple &event) {

    std::string eventName = getEventTripleName(event, h, g);
    obsTest.InsEvent(eventName);
    obsTest.SetTransition(getStateTripleName(state, h, g), eventName, "dead");
}

// if transition fobs(state,event) = newState does not exists
// adds transition into transition set of ObsTest(H, G)
// ads newState to queue to process
void newTrans(const Generator &h,
              const Generator &g,
              Generator &obsTest,
              const StateTriple &state,
              const EventTriple &event,
              std::queue<StateTriple> &queue) {

    std::string eventName = getEventTripleName(event, h, g);
    obsTest.InsEvent(eventName);

    StateTriple newState = getNextState(state, event, h, g);
    std::string newStateName = getStateTripleName(newState, h, g);

    if (!obsTest.ExistsState(newStateName)) {
        obsTest.InsState(newStateName);
        queue.push(newState);
    }

    obsTest.SetTransition(getStateTripleName(state, h, g), eventName, newStateName);
}

// isObservable() function
bool isObservable(Generator h,
                  Generator g,
                  Generator &obsTest,
                  const EventSet &obsEvents,
                  const EventSet &conEvents) {

    // check input conditions
    checkConditions(h, g, obsEvents, conEvents);

    // set up
    Idx epsilonIdx;
    std::queue<StateTriple> stateQueue;
    StateTriple state;
    EventSet actEventSet;
    setUp(h, g, obsTest, epsilonIdx, stateQueue);

    bool result = true;

    while (!stateQueue.empty()) {

        state = stateQueue.front();
        stateQueue.pop();

        // divide active event set of state triple into
        // observable and unobserble events
        actEventSet = getActEventSet(state, h, g);
        EventSet obsActEvents = actEventSet * obsEvents;
        EventSet unobsActEvents = actEventSet - obsActEvents;
        unobsActEvents.Erase(epsilonIdx);

        EventSet::Iterator it;

        // process observable events
        for (it = obsActEvents.Begin(); it != obsActEvents.End(); ++it) {

            Idx eventIdx = *it;
            if (h.ExistsTransition(state.x1, eventIdx)
                    && g.ExistsTransition(state.x3, eventIdx)) {
                if (conEvents.Exists(eventIdx)
                        &&!h.ExistsTransition(state.x2, eventIdx)) {
                    // fh(x1, e)! && fh(x2, e)not! && fg(x3, e)!
                    // dead state accessible -> K is not observable
                    EventTriple event = createEventTriple(eventIdx, epsilonIdx, eventIdx);
                    newTransToDead(h, g, obsTest, state, event);
                    result = false;
                } else {
                    // fh(x1, e)! && fh(x2, e)! && fg(x3, e)!
                    EventTriple event = createEventTriple(eventIdx, eventIdx, eventIdx);
                    newTrans(h, g, obsTest, state, event, stateQueue);
                }
            }
        }

        // process unobservable
        for (it = unobsActEvents.Begin(); it != unobsActEvents.End(); ++it) {

            Idx eventIdx = *it;
            if (h.ExistsTransition(state.x1, eventIdx)) {
                if (conEvents.Exists(eventIdx)
                        && !h.ExistsTransition(state.x2, eventIdx)
                        && g.ExistsTransition(state.x3, eventIdx)) {
                    // fh(x1, e)! && fh(x2, e)not! && fg(x3, e)!
                    // dead state accessible -> K is not observable
                    result = false;
                    EventTriple event = createEventTriple(eventIdx, epsilonIdx, eventIdx);
                    newTransToDead(h, g, obsTest, state, event);

                }
                // fh(x1, e)!
                EventTriple event = createEventTriple(eventIdx, epsilonIdx, epsilonIdx);
                newTrans(h, g, obsTest, state, event, stateQueue);
            }

            if (h.ExistsTransition(state.x2, eventIdx)
                    && g.ExistsTransition(state.x3, eventIdx)) {
                // fh(x2, e)! && fg(x3, e)!
                EventTriple event = createEventTriple(epsilonIdx, eventIdx, eventIdx);
                newTrans(h, g, obsTest, state, event, stateQueue);
            }
        }
    }

    if (result) {
        obsTest.DelState("dead");
    } else {
        obsTest.SetMarkedState("dead");
        Trim(obsTest);
    }
    return result;

}

} // end namespace faudes
