/** @file rob_observability.h verification of observability algorithm */

/*
 * Implementation of verification of observability algorithm introduced in 
 * C.G. Cassandras, S. Lafortune: Introduction to Discrete Event Systems, 
 * 2nd edition.
 * 
 * Algorithm verifies if K is observable with respect to M, Eo and Ec.
 * Input of a test:
 * G - generator, L(G) = M.
 * H - generator, L(H) = prefix closure of K, K is subset of M
 * Eo - set of observable events, Eo is subset of E; Euo = E - Eo. Let Po: E* -> Eo*.
 * Ec - set of controlable events, Ec is subset of E.
 * 
 * Output of a test:
 * ObsTest(H, G) - deterministic generator with transition function
 * involving event triples. Only marked state is state "dead".
 * 
 * If Lm(ObsTest(H, G)) is empty set then K is observable with respect to M, Eo and Ec.
 * 
 * Implemented by Marian Pochyba, 2018  
 */

#ifndef FAUDES_OBSERVABILITY_H
#define FAUDES_OBSERVABILITY_H

#include "corefaudes.h"

namespace faudes {

/**
 * Verification of observability algorithm.
 *
 * @param h
 *   The input generator; L(H) = prefix closure of K.
 * @param g
 *   The input generator; L(G) = M.
 * @param obsEvents
 *   The input set of observable events - Eo.
 * @param conEvents
 *   The input set of controlable events - Ec.
 *
 * @return
 *   True if K is observable with respect to M, Eo and Ec.
 *
 * @exception Exception
 *   - input generator nondeterministic (id 101)
 *   - L(H) is not subset of L(G) (id 201)
 *   - obsEvents/conEvents is not subset of input generator alhabet (id 202)
 *
 * @ingroup RelativeObservabilityPlugIn
 */
extern FAUDES_API bool isObservable(Generator h,
                                    Generator g,
                                    const EventSet &obsEvents,
                                    const EventSet &conEvents);

/**
 * Verification of observability algorithm and construction
 * of ObsTest(H, G) generator.
 *
 * @param h
 *   The input generator; L(H) = prefix closure of K.
 * @param g
 *   The input generator; L(G) = M.
 * @param obsTest
 *   The output generator ObsTest(H, G).
 * @param obsEvents
 *   The input set of observable events - Eo.
 * @param conEvents
 *   The input set of controlable events - Ec.
 *
 * @return
 *   True if K is observable with respect to M, Eo and Ec.
 *
 * @exception Exception
 *   - input generator nondeterministic (id 101)
 *   - L(H) is not subset of L(G) (id 201)
 *   - obsEvents/conEvents is not subset of input generator alhabet (id 202)
 *
 * @ingroup RelativeObservabilityPlugIn
 */
extern FAUDES_API bool isObservable(Generator h,
                                    Generator g,
                                    Generator &obsTest,
                                    const EventSet &obsEvents,
                                    const EventSet &conEvents);


} //end namespace faudes

#endif

