/** @file rob_relativeobservability.cpp verification of relative observability
 * and computation of supremal relatively observable sublanguage algorithms */

/*
 * Implementation of verification of relative observability
 * and computation of supremal relatively observable sublanguage agorithms
 * introduced in K. Cai, R. Zhang, and W. M. Wonham: Relative observability of
 * discrete-event systems and its supremal sublanguages, March 2015.
 *
 * Verification of relative observability
 * Input of a test:
 * G - generator.
 * A - nonblocking generator, Lm(A) = C.
 * H - onblocking generator, Lm(H) = K.
 * Eo - Eo - set of observable events, Eo is subset of E; Euo = E - Eo. Let Po: E* -> Eo*.
 *
 * Output of a test:
 * K is relatively observable with respect to prefix closure of C, G and Po: true/false.
 *
 * Computation of supremal relatively observable sublanguage
 * Input:
 * Same as in Verification of relative observability.
 *
 * Output:
 * Hsup - nonblocking generator, Lm(Hsup) = supremal relatively observable sublanguage
 * with respect to C, G and Po.
 *
 * Implemented by Marian Pochyba, 2018
 */

#include <string>

#include "rob_relativeobservability.h"

namespace faudes {

// (state, event) pair
typedef std::pair<Idx, Idx> SEPair;

// renaming function R of event e:
// e is observable -> R(e) = e-r
// e is not observable -> R(e) = e
std::string rename(const Idx &eventIdx,
                   const Generator &gen,
                   const EventSet &obsEvents) {

    if (obsEvents.Exists(eventIdx)) {
        return gen.EventName(eventIdx);
    }

    return gen.EventName(eventIdx) + "-r";
}

// step1, constructs generator Gm by marking all states of G
Generator constructGenGm(const Generator &g) {

    Generator gm = g;
    gm.Name("Gm");

    StateSet::Iterator it;
    for (it = gm.StatesBegin(); it != gm.StatesEnd(); ++it) {
        gm.SetMarkedState(*it);
    }

    return gm;
}

// step2, constructs generator M from generator A by inserting state "dead"
// and adding transitions to state "dead" when fa(x, e)not!
Generator constructGenM(const Generator &a) {

    Generator m = a;
    m.Name("M");
    m.InsState("dead");
    m.InsMarkedStates(m.States());
    Idx deadStateIdx = m.StateIndex("dead");

    StateSet::Iterator stateIt;
    EventSet::Iterator eventIt;
    for (stateIt = a.StatesBegin(); stateIt != a.StatesEnd(); ++stateIt) {
        for (eventIt = a.AlphabetBegin(); eventIt != a.AlphabetEnd(); ++eventIt) {
            if (!a.ExistsTransition(*stateIt, *eventIt)) {
                // fa(x, e)not!
                m.SetTransition(*stateIt, *eventIt, deadStateIdx);
            }
        }
    }

    return m;
}
// step3, constructs generator Mg = M x Gm
Generator constructGenMg(const Generator &m, const Generator &gm) {

    Generator mg;
    Product(m, gm, mg);
    mg.Name("Mg");

    return mg;
}

// step4, constructs generator N from generator H by clearing marked states,
// inserting marked states "dead1" and "dead2" and inserting transitions:
// 1. if fh(x,e)not! -> if e is unobservable -> fn(x,e) = dead1,
//                   -> if e is observable -> fn(x,e) = dead2.
// 2. for all e in E -> if e is unobservable -> fn(dead1, e) = dead1.
//                   -> if e is observable -> fn(dead1, e ) = dead2.
Generator constructGenN(const Generator &h, const EventSet &obsEvents) {

    Generator n = h;
    n.Name("N");
    n.ClearMarkedStates();
    n.InsMarkedState("dead1");
    n.InsMarkedState("dead2");

    Idx dead1Idx = n.StateIndex("dead1");
    Idx dead2Idx = n.StateIndex("dead2");

    StateSet::Iterator stateIt;
    EventSet::Iterator eventIt;
    for (stateIt = h.StatesBegin(); stateIt != h.StatesEnd(); ++stateIt) {
        for (eventIt = h.AlphabetBegin(); eventIt != h.AlphabetEnd(); ++eventIt) {
            if (!h.ExistsTransition(*stateIt, *eventIt)) {
                // fh(x, e)not!
                if (obsEvents.Exists(*eventIt)) {
                    // e is observable
                    n.SetTransition(*stateIt, *eventIt, dead2Idx);
                    n.SetTransition(dead1Idx, *eventIt, dead2Idx);
                } else {
                    // e is not observable
                    n.SetTransition(*stateIt, *eventIt, dead1Idx);
                    n.SetTransition(dead1Idx, *eventIt, dead1Idx);
                }
            }
        }
    }

    return n;
}

// step5, constructs generator Hc := CoAc(Mg x N)
Generator constructGenHc(const Generator &mg, const Generator &n) {

    Generator hc;
    Product(mg, n, hc);
    hc.Coaccessible();
    hc.Name("Hc");
    hc.SetDefaultStateNames();

    return hc;
}

// step6, constructs generator Hrm from generator H by applying renaming
// function R to event set, where fr(x, R(e)) = fh(x, e) and marking all states
Generator constructGenHrm(const Generator &h, const EventSet &obsEvents) {

    Generator hrm = h;
    hrm.Name("Hrm");

    StateSet::Iterator stateIt;
    for (stateIt = hrm.StatesBegin(); stateIt != hrm.StatesEnd(); ++stateIt) {
        hrm.SetMarkedState(*stateIt);
    }

    EventSet alphabet = hrm.Alphabet();
    EventSet::Iterator eventIt;
    for (eventIt = alphabet.Begin(); eventIt != alphabet.End(); ++eventIt) {
        hrm.EventRename(*eventIt, rename(*eventIt, h, obsEvents));
    }

    hrm.SetDefaultStateNames();
    return hrm;
}


// step7, constructs the verifier generator V := Hrm || Hc
Generator constructVerifier(const Generator &g,
                            const Generator &a,
                            const Generator &h,
                            const EventSet &obsEvents) {

    // step1
    Generator gm = constructGenGm(g);

    // step2
    Generator m = constructGenM(a);

    // step3
    Generator mg = constructGenMg(m, gm);

    // step4
    Generator n = constructGenN(h, obsEvents);

    // step5
    Generator hc = constructGenHc(mg, n);

    // step6
    Generator hrm = constructGenHrm(h, obsEvents);

    Generator verifier;
    Parallel(hrm, hc, verifier);
    verifier.Name("Verifier");

    return verifier;
}

// returns set of predecessor states of marked states of generator
StateSet predecessorStatesOfMarked(const Generator& gen) {

    StateSet markedStates = gen.MarkedStates();
    StateSet predecessorStates;
    StateSet::Iterator stateIt;

    for (stateIt = gen.StatesBegin(); stateIt != gen.StatesEnd(); ++stateIt) {
        if (!(gen.SuccessorStates(*stateIt) * markedStates).Empty()) {
            predecessorStates.Insert(*stateIt);
        }
    }

    return predecessorStates;
}



void checkConditions(const Generator &g,
                     const Generator &a,
                     const Generator &h,
                     const EventSet &obsEvents) {

    if (!IsDeterministic(h)) {
        std::stringstream errstr;
        errstr << "Generator " << h.Name() << " not deterministic";
        throw Exception("RelativeObservability", errstr.str(), 101);
    }

    if (!IsDeterministic(g)) {
        std::stringstream errstr;
        errstr << "Generator " << g.Name() << " not deterministic";
        throw Exception("RelativeObservability", errstr.str(), 101);
    }

    if (!IsDeterministic(a)) {
        std::stringstream errstr;
        errstr << "Generator " << a.Name() << " not deterministic";
        throw Exception("RelativeObservability", errstr.str(), 101);
    }

    if (!IsNonblocking(a)) {
        std::stringstream errstr;
        errstr << "Generator " << a.Name() << " must be nonblocking, but is blocking!";
        throw Exception("RelativeObservability", errstr.str(), 201);
    }

    if (!IsNonblocking(h)) {
        std::stringstream errstr;
        errstr << "Generator " << h.Name() << " must be nonblocking, but is blocking!";
        throw Exception("RelativeObservability", errstr.str(), 201);
    }

    if (!LanguageInclusion(h, a)) {
        std::stringstream errstr;
        errstr << "Lm(H) = K is not a subset of Lm(A) = C!";
        throw Exception("Observability", errstr.str(), 202);
    }

    if (!LanguageInclusion(a, g)) {
        std::stringstream errstr;
        errstr << "Lm(A) = C is not a subset of Lm(G)!";
        throw Exception("Observability", errstr.str(), 202);
    }

    EventSet sigma = g.Alphabet();
    if (!(obsEvents <= sigma)) {
        std::stringstream errstr;
        errstr << "Observable event set is not a subset of Alphabet of G!";
        throw Exception("Observability", errstr.str(), 203);
    }

}
// isRelativelyObservable() function
bool isRelativelyObservable(const Generator &g,
                            const Generator &a,
                            const Generator &h,
                            const EventSet &obsEvents) {

    // check input conditions
    checkConditions(g, a, h, obsEvents);

    // step1 - step7
    Generator verifier = constructVerifier(g, a, h, obsEvents);


    StateSet predecStates = predecessorStatesOfMarked(verifier);
    StateSet markedStates = verifier.MarkedStates();
    StateSet::Iterator stateIt;
    EventSet::Iterator eventIt;

    // step8 - for all (x, e) that fv(x, e) transitions into marked states of verifier
    // checks whether one of following conditions hold true:
    // e is in Eo
    // e is not in Eo && fv(x, R(e))not!
    for (stateIt = predecStates.Begin(); stateIt != predecStates.End(); ++stateIt) {
        EventSet actEventSet = verifier.ActiveEventSet(*stateIt);
        for (eventIt = actEventSet.Begin(); eventIt != actEventSet.End(); eventIt++) {
            if (!(verifier.SuccessorStates(*stateIt, *eventIt) * markedStates).Empty()) {
                if (obsEvents.Exists(*eventIt)
                        || (!obsEvents.Exists(*eventIt)
                            && verifier.ExistsTransition(*stateIt, verifier.EventIndex(rename(*eventIt,
                                                                                              verifier,
                                                                                              obsEvents))))) {
                        return false;
                }
            }
        }
    }

    return true;
}
// finds and stores parts (pairs (x, e)) of generator Hs which brekas the property of
// relative observability
void getNotRelObsParts(const Generator &verifier,
                       const Generator &hs,
                       std::set<SEPair> &pairSet,
                       const EventSet &obsEvents) {

    StateSet predecStates = predecessorStatesOfMarked(verifier);
    StateSet markedStates = verifier.MarkedStates();

    StateSet::Iterator stateIt;
    EventSet::Iterator eventIt;

    for (stateIt = predecStates.Begin(); stateIt != predecStates.End(); ++stateIt) {
        EventSet actEventSet = verifier.ActiveEventSet(*stateIt);
        for (eventIt = actEventSet.Begin(); eventIt != actEventSet.End(); eventIt++) {
            if (!(verifier.SuccessorStates(*stateIt, *eventIt) * markedStates).Empty()) {
                if (obsEvents.Exists(*eventIt)
                        || (!obsEvents.Exists(*eventIt)
                            && verifier.ExistsTransition(*stateIt, verifier.EventIndex(rename(*eventIt,
                                                                                              verifier,
                                                                                              obsEvents))))) {
                    SEPair pair;
                    std::string stateName = verifier.StateName(*stateIt);

                    // store index of "not observable state" in generator Hs
                    // stateName in Hs is first half of stateName in verifier generator, divided by "|"
                    pair.first = hs.StateIndex(stateName.substr(0, stateName.find("|", 0)));

                    // store index of "not observable event"
                    pair.second = *eventIt;
                    pairSet.insert(pair);

                }
            }
        }
    }
}

// cleares parts (pairs (x, e)) of generator Hs which brekas the property of
// relative observability
void clrNotRelObsParts(Generator& hs, const std::set<SEPair> &pairSet) {

    std::set<SEPair>::iterator it;
    for (it = pairSet.begin(); it != pairSet.end(); ++it) {
        SEPair pair = *it;
        hs.ClrTransitions(pair.first, pair.second);
    }

    Trim(hs);
}

// finds and cleares parts (pairs (x, e)) of generator Hs which brekas the property of
// relative observability
void getAndClrNotRelObsParts(const Generator &g,
                         const Generator &a,
                         const Generator &h,
                         Generator &hs,
                         const EventSet &obsEvents) {

    bool repeat = true;
    while (repeat) {

        // step3 construct verifier generator from G, A, Hs, Eo
        Generator verifier = constructVerifier(g, a, hs, obsEvents);

        // step4 if verifier is not empty form the set of parts(pairs (x, e))
        // which breaks the property of  relative observability
        std::set<SEPair> pairSet;
        if (!verifier.Empty()) {
            getNotRelObsParts(verifier, hs, pairSet, obsEvents);
        }

        // step5
        //   - 5.1 for all (x, e) exclude transition (x, e, fhs(x, e)) from generator Hs
        //   - 5.2 Trim(Hs)
        //   - 5.3 return to step3
        if (!pairSet.empty()) {
            clrNotRelObsParts(hs, pairSet);
        } else {
            repeat = false;
        }
        pairSet.clear();
    }
}

// supRelativelyObservable() function
void supRelativelyObservable(const Generator &g,
                             const Generator &a,
                             const Generator &h,
                             Generator &result,
                             const EventSet &obsEvents) {

    // check input conditions
    checkConditions(g, a, h, obsEvents);

    // step1, construct generator Hsp := H || Obs(H, Eo)
    // Obs(H, Eo) - observer of H w.r.t. Eo
    Generator hsp;
    Generator obs;
    Project(h, obsEvents, obs);
    Parallel(h, obs, hsp);
    hsp.Name("Hsp");
    hsp.SetDefaultStateNames();
    // step2 Hs := Hsp
    Generator hs = hsp;
    hs.Name("Hs");

    // step3 - step5
    getAndClrNotRelObsParts(g, a, h, hs, obsEvents);

    // step6 Hsup := Hs
    result = hs;
    result.Name("Hsup");
}

} // end namespace faudes
