/** @file rob_relativeobservability.h verification of relative observability
 * and computation of supremal relatively observable sublanguage algorithms */

/*
 * Implementation of verification of relative observability
 * and computation of supremal relatively observable sublanguage agorithms
 * introduced in K. Cai, R. Zhang, and W. M. Wonham: Relative observability of
 * discrete-event systems and its supremal sublanguages, March 2015.
 *
 * Verification of relative observability
 * Input of a test:
 * G - generator.
 * A - nonblocking generator, Lm(A) = C.
 * H - onblocking generator, Lm(H) = K.
 * Eo - Eo - set of observable events, Eo is subset of E; Euo = E - Eo. Let Po: E* -> Eo*.
 *
 * Output of a test:
 * K is relatively observable with respect to prefix closure of C, G and Po: true/false.
 *
 * Computation of supremal relatively observable sublanguage
 * Input:
 * Same as in Verification of relative observability.
 *
 * Output:
 * Hsup - nonblocking generator, Lm(Hsup) = supremal relatively observable sublanguage
 * with respect to C, G and Po.
 *
 * Implemented by Marian Pochyba, 2018
 */

#ifndef FAUDES_RELATIVEOBS_H
#define FAUDES_RELATIVEOBS_H

#include "corefaudes.h"

namespace faudes {

/**
 * Verification of relative observability algorithm.
 *
 * @param g
 *   The input generator G.
 * @param a
 *   The input nonblocking generator A; Lm(A) = C.
 * @param h
 *   The input nonblocking generator H; Lm(H) = K.
 * @param obsEvents
 *   The input set of observable events.
 *
 * @return
 *   True if K is relatively observable with respect to prefix closure of C, G and Po.
 *
 * @exception Exception 
 *   - input generator non-deterministic (id 101)
 *   - generator A or H is blocking (id 201)
 *   - Lm(H) is not subset of Lm(A) or Lm(A) is not subset of Lm(G) (id 202)
 *   - obsEvents is not subset of input generator alphabet (id 203)
 *
 * @ingroup RelativeObservabilityPlugIn
 */
extern FAUDES_API bool isRelativelyObservable(const Generator &g,
                           		      const Generator &a,
                                              const Generator &h,
                                              const EventSet &obsEvents);

/**
 * Computation of supremal relatively observable sublanguage algorithm.
 *
 * @param g
 *   The input generator G.
 * @param a
 *   The input nonblocking generator A; Lm(A) = C.
 * @param h
 *   The input nonblocking generator H; Lm(H) = K.
 * @param result
 *   The output nonblocking generator; Lm(result) = supremal relatively
 *   observable sublanguage with respect to C, G and Po.
 * @param obsEvents
 *   The input set of observable events.
 *
 * @exception Exception 
 *   - input generator non-deterministic (id 101)
 *   - generator A or H is blocking (id 201)
 *   - Lm(H) is not subset of Lm(A) or Lm(A) is not subset of Lm(G) (id 202)
 *   - obsEvents is not subset of input generator alphabet (id 203)
 *
 * @ingroup RelativeObservabilityPlugIn
 */
extern FAUDES_API void supRelativelyObservable(const Generator &g,
                                               const Generator &a,
                                               const Generator &h,
                                               Generator &result,
                                               const EventSet &obsEvents);

} // end namespace faudes

#endif
