/** @file rob_obs.cpp

Tutorial, supervisory control

@ingroup Tutorials 

@include rob_obs.cpp

*/

#include "libfaudes.h"

// make the faudes namespace available to our program
using namespace faudes;


// main program
int main() {

    std::cout << "Reading generators..." << std::endl;
    // read generators from files
    Generator h1;
    h1.Read("data/obs_h1.gen");
    Generator h2;
    h2.Read("data/obs_h2.gen");
    Generator h3;
    h3.Read("data/obs_h3.gen");

    Generator g1;
    g1.Read("data/obs_g.gen");
    Generator g2;
    g2.Read("data/obs_g2.gen");

    std::cout << "Reading event sets..." << std::endl;
    // read observable and controlable event sets from files
    EventSet obsEvents1;
    obsEvents1.Read("data/obs_obs_events1.alph");
    EventSet obsEvents2;
    obsEvents2.Read("data/obs_obs_events2.alph");
    EventSet obsEvents3;
    obsEvents3.Read("data/obs_obs_events3.alph");

    EventSet conEvents1;
    conEvents1.Read("data/obs_con_events1.alph");
    EventSet conEvents2;
    conEvents2.Read("data/obs_con_events2.alph");
    EventSet conEvents3;
    conEvents3.Read("data/obs_con_events3.alph");

    std::cout << "Input generators are read from directory \"data\". Visual representation are also included." << std::endl << std::endl;
    std::cout << "Example 1" << std::endl;
    std::cout << "Testing if L(H1) is observable with respect to L(G1), Eo1 = {b} and Ec1 = {b} with result:" << std::endl;


    // now we test if K1 is observable with respect to M1, Eo1 and Ec1
    // L(G1) = M1, L(H1) = prefix-closure(K1)
    // Eo1 - observable events - obsEvents1
    // Ec1 - controlable events - conEvents1

    if (isObservable(h1, g1, obsEvents1, conEvents1)) {
        std::cout << "Language L(H1) is observable with respect to L(G1), Eo1 and Ec1!" << std::endl;
    } else {
        std::cout << "Language L(H1) is not observable with respect to L(G1), Eo1 and Ec1!" << std::endl;
    }

    std::cout << std::endl;
    std::cout << "Example 2" << std::endl;
    std::cout << "Testing if L(H2) is observable with respect to L(G1), Eo2 = {b} and Ec2 = {u, b} with result:" << std::endl;

    // now we test if K2 is observable with respect to M1, Eo2 and Ec2
    // we also construct  generator ObsTest(H2, G1),
    // which records all violations of observability
    // L(G1) = M1, L(H2) = prefix-closure(K2)
    // Eo2 - observable events - obsEvents2
    // Ec2 - controlable events - conEvents2

    Generator obsTest1;
    if (isObservable(h2, g1, obsTest1, obsEvents2, conEvents2)) {
        std::cout << "Language L(H2) is observable with respect to L(G1), Eo2 and Ec2!" << std::endl;
    } else {
        std::cout << "Language L(H2) is not observable with respect to L(G1), Eo2 and Ec2!" << std::endl;
    }

    std::cout << std::endl << "Also constructed generator ObsTest(H2, G) that, records all violations of observability. "
                              "Creating a visual representation in directory \"results\"." << std::endl;

    // now we can further work with ObsTest(H2,G), e.g. create an image of it
    obsTest1.GraphWrite("results/obstest_h2_g.png");

    std::cout << std::endl;
    std::cout << "Example 3" << std::endl;
    std::cout << "Testing if L(H3) is observable with respect to L(G2), Eo3 = {a, d} and Ec3 = {a, b, c, d} with result:" << std::endl;

    // now we test if K3 is observable with respect to M2, Eo3 and Ec3
    // we also construct  generator ObsTest(H3, G2),
    // which records all violations of observability
    // L(G2) = M2, L(H3) = prefix-closure(K3)
    // Eo3 - observable events - obsEvents3
    // Ec3 - controlable events - conEvents3

    Generator obsTest2;
    if (isObservable(h3, g2, obsTest2, obsEvents3, conEvents3)) {
        std::cout << "Language L(H3) is observable with respect to L(G2), Eo3 and Ec3!" << std::endl;
    } else {
        std::cout << "Language L(H3) is not observable with respect to L(G2), Eo3 and Ec3!" << std::endl;
    }

    std::cout << std::endl << "Also constructed generator ObsTest(H3, G2) that, records all violations of observability. "
                              "Creating a visual representation in directory \"results\"." << std::endl;

    // now we can further work with ObsTest(H2,G), e.g. create an image of it
    obsTest2.GraphWrite("results/obstest_h3_g1.png");

    // done
    return 0;
}



